#define _GNU_SOURCE
#include <errno.h>
#include <libspe.h>
#include <pthread.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/mman.h>

#include "spe.h"

#define __PRINTF(fmt, args...) { fprintf(stderr,fmt , ## args); }
#ifdef DEBUG
#define DEBUG_PRINTF(fmt, args...) __PRINTF(fmt , ## args)
#else
#define DEBUG_PRINTF(fmt, args...)
#endif

#define SPE_SIGNAL_REG 0x0C

unsigned int *spe_signal_address(speid_t speid, int signo)
{
	struct thread_store *thread_store = speid;
	int fd;
	void *base_addr;

	if (signo != 1 && signo != 2)
	{
		perror("spe_signal_address: only signal 1 and 2 are valid");
		errno = EINVAL;
		return MAP_FAILED;
	}

	if (signo == 1)
	{
		fd = thread_store->fd_sig1;
		base_addr = thread_store->signal1_mmap_base;
	}
	else
	{
		fd = thread_store->fd_sig2;
		base_addr = thread_store->signal2_mmap_base;
	}


	if ( base_addr == MAP_FAILED )
	{	
		base_addr=mmap(0, 0x1000, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
		
		if ( base_addr == MAP_FAILED )
		{	
			perror("spe_signal_address: internal error. cant map signal file.");
			errno = EIO;
			return MAP_FAILED;
		}
	}


	return base_addr+0x0C;
}

