/*
 * libspe - A wrapper library to adapt the JSRE SPU usage model to SPUFS
 * Copyright (C) 2005 IBM Corp.
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License,
 * or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 *  License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software Foundation,
 *   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <elf.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#define _GNU_SOURCE
#include <sched.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <wait.h>

#include <linux/unistd.h>

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/syscall.h>

#include <sys/spe.h>

#include "libspe.h"
#include "spe.h"
#include "elf_loader.h"
#include "spe_exec.h"
#include "default_c99_handler.h"
#include "default_posix1_handler.h"


#define __PRINTF(fmt, args...) { fprintf(stderr,fmt , ## args); }
#ifdef DEBUG
#define DEBUG_PRINTF(fmt, args...) __PRINTF(fmt , ## args)
#else
#define DEBUG_PRINTF(fmt, args...)
#endif

/* More declarations
 */

#ifdef HAVE_GLIBC23
long sched_setaffinity(pid_t pid, unsigned int len, unsigned long *user_mask_ptr);
long sched_getaffinity(pid_t pid, unsigned int len, unsigned long *user_mask_ptr);
pid_t gettid(void);
_syscall0(pid_t,gettid)
#else
int pthread_attr_setaffinity_np (pthread_attr_t *__attr, size_t __cpusetsize, __const cpu_set_t *__cpuset);
int pthread_setaffinity_np (pthread_t __th, size_t __cpusetsize, __const cpu_set_t *__cpuset);
int pthread_getaffinity_np (pthread_t __th, size_t __cpusetsize, cpu_set_t *__cpuset);
#endif

/*
 * Helpers
 *
 * */

static struct group_list grp_list = {NULL,PTHREAD_MUTEX_INITIALIZER,0};

void env_check(void);
int __env_spu_debug_start = 0;
int __env_spu_info = 0;
int check_env = 1;


/* Group defaults
 *  */

int default_priority = 0;
int default_policy   = SCHED_OTHER;
int default_eventmask = 0;


/*Default SPE library call handlers for 21xx stop-and-signal.
*/
static void *handlers[] = {
	default_c99_handler, default_posix1_handler,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,
	NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL };

/*
 * int spe_ldr[]:
 * SPE code that performs the actual parameter setting:
 */
static int spe_ldr[] = {
	0x30fff203,
	0x30fff404,
	0x30fff605,
	0x30fff806,
	0x30fff000,
	0x35000000,
	0x00000000,
	0x00000000
};

void
env_check(void)
{
  if (getenv("SPU_DEBUG_START"))
    {
      __env_spu_debug_start = strtol ((const char *)getenv("SPU_DEBUG_START"),(char **)NULL, 0);
    }
  if (getenv("SPU_INFO"))
    {
      __env_spu_info = strtol ((const char *)getenv("SPU_INFO"),(char **)NULL, 0);
    }
  check_env = 0;
}


static int
spe_open_files(struct thread_store *spe, const char *pathname)
{
	char filename[256];
	int ret;
	
	sprintf(filename, "%s/mbox", pathname);
	spe->fd_mbox = open(filename, O_RDONLY);

	sprintf(filename, "%s/mbox_stat", pathname);
	spe->fd_mbox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/ibox", pathname);
	spe->fd_ibox = open(filename, O_RDONLY);

	sprintf(filename, "%s/ibox_stat", pathname);
	spe->fd_ibox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/wbox", pathname);
	spe->fd_wbox = open(filename, O_WRONLY);

	sprintf(filename, "%s/wbox_stat", pathname);
	spe->fd_wbox_stat = open(filename, O_RDONLY);

	sprintf(filename, "%s/npc", pathname);
	spe->fd_npc = open(filename, O_RDWR);
		
	sprintf(filename, "%s/signal1", pathname);
	spe->fd_sig1 = open(filename, O_RDWR);
	spe->signal1_mmap_base=MAP_FAILED;
	
	sprintf(filename, "%s/signal2", pathname);
	spe->fd_sig2 = open(filename, O_RDWR);
	spe->signal2_mmap_base=MAP_FAILED;
	
	/* this one may fail */
	sprintf(filename, "%s/mfc", pathname);
	spe->fd_mfc = open(filename, O_RDWR);
	spe->mfc_mmap_base=MAP_FAILED;

	/* this one may fail */
	sprintf(filename, "%s/cntl", pathname);
	spe->fd_cntl = open(filename, O_RDWR);
	spe->cntl_mmap_base=MAP_FAILED;

/*
 	if (spe->fd_cntl != -1)
	{
		spe->cntl_mmap_base = mmap(0, 0x2000, PROT_READ | PROT_WRITE, MAP_SHARED, spe->fd_cntl, 0);
	}
*/
	ret=pipe(spe->ev_pipe);
		
	if (spe->fd_mbox      < 0 || spe->fd_ibox      < 0 || spe->fd_wbox      < 0 ||
	    spe->fd_mbox_stat < 0 || spe->fd_ibox_stat < 0 || spe->fd_wbox_stat < 0 ||
	    spe->fd_npc       < 0 || spe->fd_sig1      < 0 || spe->fd_sig2      < 0 || 
	    ret ) {
		DEBUG_PRINTF("Could not open all SPE files.\n");
		close(spe->fd_mbox);
		close(spe->fd_mbox_stat);
		close(spe->fd_ibox);
		close(spe->fd_ibox_stat);
		close(spe->fd_wbox);
		close(spe->fd_wbox_stat);
		close(spe->fd_npc);
		close(spe->fd_sig1);
		close(spe->fd_sig2);
		close(spe->fd_mfc);
/*
		if (spe->cntl_mmap_base != MAP_FAILED)
		{
			munmap(spe->cntl_mmap_base,0x2000);
		}
*/
		close(spe->fd_cntl);
		close(spe->ev_pipe[0]);
		close(spe->ev_pipe[1]);
		return -1;
	}
	return 0;
}


/**
 * Low-level API
 * 
 */
int 
add_group_to_groups(struct group_store *gid)
{
	struct grpsElem *listElem;
	listElem = malloc(sizeof *listElem);

	if (!listElem)
	{
		errno=ENOMEM;
		return -errno;
	}
	
	pthread_mutex_lock(&grp_list.mutex);

	listElem->group=gid;
	listElem->next=grp_list.list_start;
	
	grp_list.list_start=listElem;
	grp_list.numListSize++;
		
	pthread_mutex_unlock(&grp_list.mutex);
	
	return 0;
	
}

int 
add_thread_to_group(struct group_store *gid, struct thread_store *thread)
{
	struct grpListElem *listElem;
	listElem = malloc(sizeof *listElem);

	if (!listElem)
	{
		errno=ENOMEM;
		return -errno;
	}

	pthread_mutex_lock(&grp_list.mutex);
	
	listElem->thread=thread;
	listElem->next=gid->grp_members;

	gid->grp_members=listElem;
	gid->numListSize++;
		
	pthread_mutex_unlock(&grp_list.mutex);
	
	return 0;
}

int 
remove_thread_from_group(struct group_store *gid, struct thread_store *thread)
{
	pthread_mutex_lock(&grp_list.mutex);
	
	if (gid->numListSize == 1)
	{	
		struct grpListElem *p=gid->grp_members;
		gid->numListSize--;
		free(p);
		gid->grp_members = NULL;
		pthread_mutex_unlock(&grp_list.mutex);
		return 0;
	}

	for (struct grpListElem *p=gid->grp_members; p->next!=NULL; p=p->next)
	{
		struct grpListElem *np=p->next;
		if (np->thread==thread)
		{
			p->next=np->next;
			gid->numListSize--;
			free(np);
			
			pthread_mutex_unlock(&grp_list.mutex);

			return 0;
		}
	}

	pthread_mutex_unlock(&grp_list.mutex);

	errno = ESRCH;
	return -ESRCH;
}

struct thread_store *
srch_thread(struct thread_store *thread)
{
	pthread_mutex_lock(&grp_list.mutex);

	for (struct grpsElem *gp=grp_list.list_start ; gp !=NULL; gp=gp->next)
	{
		//printf("1\n");
		struct group_store *gid = gp->group;
			
		for (struct grpListElem *p=gid->grp_members; p !=NULL; p=p->next)
		{
			//printf("2\n");
			if (p->thread==thread)
			{
				pthread_mutex_unlock(&grp_list.mutex);		
				return p->thread;
			}
		}
	}
	
	pthread_mutex_unlock(&grp_list.mutex);
	errno = ESRCH;
	return NULL;
}

struct group_store *
srch_group(struct group_store *group)
{
	pthread_mutex_lock(&grp_list.mutex);

	for (struct grpsElem *gp=grp_list.list_start ; gp !=NULL; gp=gp->next)
	{
		if (gp->group == group)
		{
			pthread_mutex_unlock(&grp_list.mutex);		
			return gp->group;
		}
	}
	
	pthread_mutex_unlock(&grp_list.mutex);
	errno = ESRCH;
	return NULL;
}


void *
spe_gid_setup(int policy, int priority, int use_events)
{
	struct group_store *group_store;

        group_store = calloc (1, sizeof *group_store);
	if (!group_store)
	{
		DEBUG_PRINTF ("Could not allocate group store\n");
		errno = ENOMEM;
	        return NULL;
	}
	
	group_store->policy = policy;
	group_store->priority = priority;
	group_store->use_events = use_events;
	group_store->grp_members = NULL;
	group_store->numListSize = 0;

	add_group_to_groups(group_store);
	
	return group_store;
}

void *
spe_setup (spe_gid_t gid, spe_program_handle_t *handle, void *argp, void *envp, int flags)
{
	struct thread_store *thread_store;
	int rc, memfd, sigfd;
	addr64 argp64, envp64, tid64;
	char signame[256], memname[256], pathname[256];
	void *spe_ld_buf;
#ifdef SPE_NO_MMAP
	ssize_t count = 0, num = 0;
#endif
	struct spe_ld_info ld_info;
	struct spe_exec_params spe_params __attribute__ ((aligned (4096)));
	
	if(!gid)
	{
		gid=spe_gid_setup(default_policy,default_priority,default_eventmask);
	}
	else
	{
		struct group_store *grp = gid;
		if (grp->numListSize == MAX_THREADS_PER_GROUP)
		{
		  	DEBUG_PRINTF ("Thread group has reached maximum number of members.\n");
		  	errno = ENOMEM;
		  	return NULL;
			
		}
	}

	/*Allocate thread structure */

	thread_store = calloc (1, sizeof *thread_store);
	if (!thread_store)
	  {
		  DEBUG_PRINTF ("Could not allocate thread store\n");
		  errno = ENOMEM;
		  return NULL;
	  }

	argp64.ull = (unsigned long long) (unsigned long) argp;
	envp64.ull = (unsigned long long) (unsigned long) envp;

	thread_store->group_id = gid;
	thread_store->flags = flags;

	tid64.ull = (unsigned long long) (unsigned long) thread_store;
	
	/* Make the SPU Directory */

	sprintf (pathname, "/spu/spethread-%i-%lu",
		 getpid (), (unsigned long)thread_store);

	DEBUG_PRINTF ("create %s\n", pathname);

	rc = spe_create(pathname, 0, S_IRUSR | S_IWUSR | S_IXUSR);
	thread_store->fd_run = rc;
	if (rc < 0)
	  {
		  DEBUG_PRINTF ("Could not create SPE %s\n", pathname);
		  return NULL;
	  }

	sprintf (memname, "%s/mem", pathname);

	/* Check SPE */

	memfd = open (memname, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
	if (memfd < 0)
	  {
		  DEBUG_PRINTF ("Could not open SPE mem file.\n");
		  return NULL;
	  }

	/* Prepare Loader */
#ifdef SPE_NO_MMAP	
	spe_ld_buf = malloc (LS_SIZE);
#else
	spe_ld_buf = mmap(0, LS_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, 0);
	thread_store->mem_mmap_base = spe_ld_buf;
	thread_store->fd_mem = memfd;
#endif
	if (!spe_ld_buf)
	  {
		  DEBUG_PRINTF ("Could not allocate SPE memory. \n");
		  errno = ENOMEM;
		  return NULL;
	  }

	memset(spe_ld_buf, 0, LS_SIZE);

	rc = load_spe_elf (handle, spe_ld_buf, &ld_info);
	if (rc != 0)
	  {
		  DEBUG_PRINTF ("Load SPE ELF failed..\n");
		  return NULL;
	  }

	thread_store->handle = *handle;

	/* Add SPE exec program */

	DEBUG_PRINTF ("Add exec prog dst:0x%04x size:0x%04x\n",
		      SPE_LDR_START, sizeof (spe_ldr));
	memcpy (spe_ld_buf + SPE_LDR_START, &spe_ldr, sizeof (spe_ldr));

	/* Add SPE exec parameters */

	spe_params.gpr0[0] = ld_info.entry;
	spe_params.gpr0[1] = 0;
	spe_params.gpr0[2] = 0;
	spe_params.gpr0[3] = 0;
	
	if (flags & SPE_USER_REGS) {
		/* When flags & SPE_USER_REGS is set, argp points
		 * to an array of 3x128b registers to be passed
		 * directly to the SPE program.
		 */
		memcpy(&spe_params.gpr3[0], argp, sizeof(unsigned int) * 12);
	} else {
		/* Regular parameter convention for spe_create_thread. */
		spe_params.gpr3[0] = tid64.ui[0];
		spe_params.gpr3[1] = tid64.ui[1];
		spe_params.gpr3[2] = 0;
		spe_params.gpr3[3] = 0;
	
		spe_params.gpr4[0] = argp64.ui[0];
		spe_params.gpr4[1] = argp64.ui[1];
		spe_params.gpr4[2] = 0;
		spe_params.gpr4[3] = 0;
	
		spe_params.gpr5[0] = envp64.ui[0];
		spe_params.gpr5[1] = envp64.ui[1];
		spe_params.gpr5[2] = 0;
		spe_params.gpr5[3] = 0;
	}

	spe_params.gpr6[0] = 0;
	spe_params.gpr6[1] = 0;
	spe_params.gpr6[2] = 0;
	spe_params.gpr6[3] = 0;

	DEBUG_PRINTF ("Add exec param dst:0x%04x size:0x%04x\n",
		      SPE_PARAM_START, sizeof (spe_params));
	memcpy (spe_ld_buf + SPE_PARAM_START, &spe_params,
		sizeof (spe_params));

#ifdef SPE_NO_MMAP
	/* Copy SPE image to SPUfs */
	do
	  {
		  num = write (memfd, spe_ld_buf + count, LS_SIZE - count);
		  if (num == -1)
		    {
			    DEBUG_PRINTF ("Transfer SPE ELF failed..\n");
			    return NULL;
		    }

		  count += num;
	  }
	while (count < LS_SIZE && num);
	close (memfd);

	/* Free the SPE Buffer */
	free (spe_ld_buf);
#endif

	rc = spe_open_files(thread_store, pathname);
	if (rc) {
		perror("open files\n");
		return NULL;
	}

	rc = add_thread_to_group(gid, thread_store);
        
	if ( flags & SPE_CFG_SIGNOTIFY1_OR )
        {
                char one[] = "1";
		int ignore;

                sprintf (signame, "%s/signal1_type", pathname);
                sigfd = open (signame, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
                if (sigfd < 0)
                {
                        DEBUG_PRINTF ("Could not open SPE signal1_type file.\n");
                        return NULL;
                }
                ignore=write (sigfd, one, 1);
                close(sigfd);
        }
        if ( flags & SPE_CFG_SIGNOTIFY2_OR )
        {
                char one[] = "1";
		int ignore;
   
                sprintf (signame, "%s/signal2_type", pathname);
                sigfd = open (signame, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
                if (sigfd < 0)
                {
                        DEBUG_PRINTF ("Could not open SPE signal2_type file.\n");
                        return NULL;
                } 
                ignore=write (sigfd, one, 1);
                close(sigfd); 
        }

	sync();

	thread_store->stop = 0;

	return thread_store;
}

void
spe_cleanup (void *ptr)
{
	struct thread_store *thread_store = ptr;
	char pathname[256];

	sprintf (pathname, "/spu/spethread-%i-%lu",
		 getpid (), (unsigned long)thread_store);

	close (thread_store->fd_run);
	close (thread_store->fd_mbox);
	close (thread_store->fd_mbox_stat);
	close (thread_store->fd_ibox);
	close (thread_store->fd_ibox_stat);
	close (thread_store->fd_wbox);
	close (thread_store->fd_wbox_stat);
	close (thread_store->fd_npc);
	
	if (thread_store->signal1_mmap_base != MAP_FAILED)
	{
		munmap(thread_store->signal1_mmap_base,0x1000);
	}
	close (thread_store->fd_sig1);
	
	if (thread_store->signal2_mmap_base != MAP_FAILED)
	{
		munmap(thread_store->signal2_mmap_base,0x1000);
	}
	close (thread_store->fd_sig2);
	
	if (thread_store->mfc_mmap_base != MAP_FAILED)
	{
		munmap(thread_store->mfc_mmap_base,0x1000);
	}
	close (thread_store->fd_mfc);
	
	if (thread_store->cntl_mmap_base != MAP_FAILED)
	{
		munmap(thread_store->cntl_mmap_base,0x2000);
	}
	close (thread_store->fd_cntl);

#ifndef SPE_NO_MMAP
	munmap(thread_store->mem_mmap_base,LS_SIZE);
	close (thread_store->fd_mem);
#endif
	close (thread_store->ev_pipe[0]);
        close (thread_store->ev_pipe[1]);

#ifndef HAVE_GLIBC23
	pthread_attr_destroy(&thread_store->thread_attr);	
#endif	

	remove_thread_from_group(thread_store->group_id, thread_store);

	free (thread_store);
}

unsigned int
set_npc (void *ptr, unsigned int npc)
{
	struct thread_store *thread_store = ptr;
	int ret;
	char npcText[9];

	sprintf(npcText,"0x%06x\n",npc);
	ret = write(thread_store->fd_npc, npcText, 9);
	if (ret < 0)
        {
                return -EINVAL;
        }
	return npc;
}

int
do_spe_run (void *ptr)
{
	struct thread_store *thread_store = ptr;
	int runfd;

	unsigned int npc;
	unsigned int code;
	unsigned int status;

	npc = SPE_LDR_PROG_start;

	int ret, rc;

        DEBUG_PRINTF ("do_spe_run\n");
	runfd = thread_store->fd_run;

#ifdef HAVE_GLIBC23
	//record our tid
	thread_store->threadid = gettid();
	sched_setaffinity(thread_store->threadid, sizeof(unsigned long), &thread_store->mask);
#endif
	thread_store->thread_status = 1; 
	
	/* Note: in order to be able to kill a running SPE thread the 
	 *       spe thread must be cancelable.
	 */
	
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);
	
	do {
		thread_store->thread_status = 2; 
		
		ret = spe_run(runfd, &npc, &status);
		code = ret >> 16;
                DEBUG_PRINTF ("SPE (npc=0x%06x) (code=0x%04x) (ret=0x%04x).\n",
			      npc, code, ret & 0xffff);

		thread_store->thread_status = 10; //SPE _not_ running.
		
		/* If a signal was sent to STOP execution we will pause and wait 
		 * for a continue signal before going on.
		 */

		if ( thread_store->stop )
		{
			pthread_mutex_lock(&thread_store->event_lock);
			pthread_cond_init(&thread_store->event_deliver, NULL);
			
			thread_store->thread_status = 20; //SPE _pausing_.
			/* Now wait for the continue signal */
			
			pthread_cond_wait(&thread_store->event_deliver, &thread_store->event_lock);
			pthread_mutex_unlock(&thread_store->event_lock);
			thread_store->stop = 0;
			thread_store->thread_status = 10; //SPE _not_ running.
		}
                
		if (ret < 0) {
                          DEBUG_PRINTF ("Could not run on SPE: %s.\n",
				strerror(errno));
			  return -errno;
                } else if ((code&0xff00)==0x2100) {
			int callnum = code&0xff;
			
			DEBUG_PRINTF ("SPE library call: ");
			
			if (!handlers[callnum])
			{
				DEBUG_PRINTF ("Unsupported SPE library call signal: 21%i.",callnum);
				break;
			}
			else
			{
				int (*handler)(void *, unsigned int);
				handler=handlers[callnum];
				
				rc = handler(thread_store->mem_mmap_base, npc);
				if (rc) 
				{
			        	DEBUG_PRINTF ("SPE library call unsupported.\n");
					thread_store->thread_status = 99; 
				        return -ENOSYS;
			        }
				else
				{
					npc += 4;
				}
			}

		}
		else if (code < 0x2000)
		{
			pthread_mutex_lock(&thread_store->event_lock);
			pthread_cond_init(&thread_store->event_deliver, NULL);
			
			if (((struct group_store*)thread_store->group_id)->use_events)
			{
				int callnum = code;
				//
				// Event - driven 21xx hanlers
				//
				
				thread_store->event = SPE_EVENT_STOP;
				thread_store->ev_data = callnum;
				//
				//Todo Error handling
				//
				write(thread_store->ev_pipe[1], &thread_store->event, 4);
				
				//Show that we're waiting
				thread_store->event_pending = 1;
				
				// Wait for a signal.
				pthread_cond_wait(&thread_store->event_deliver, &thread_store->event_lock);

				thread_store->event_pending = 0;
				
				pthread_mutex_unlock(&thread_store->event_lock);
			}
			else
			{

				pthread_mutex_unlock(&thread_store->event_lock);
				
				DEBUG_PRINTF ("SPE user events not enabled.\n");
				thread_store->thread_status = 99; 
			        return -ENOSYS;
			}
		}

        } while ((code & 0xff00) == 0x2100 || (code < 0x2000));

	DEBUG_PRINTF ("SPE thread result: %08x:%04x\n", npc, code);
	
	/* Save status & npc */	
	thread_store->npc = npc;
	thread_store->ret_status = ret;
	
	//
	//Event code
	//
	if (((struct group_store*)thread_store->group_id)->use_events)
	{
		thread_store->event = SPE_EVENT_THREAD_EXIT;
		thread_store->ev_data = code & 0xff;
		//
		//Todo Error handling
		//
		write(thread_store->ev_pipe[1], &thread_store->event, 4);
	}

	thread_store->thread_status = 99; 
	return code;
}

struct image_handle {
	spe_program_handle_t speh;
	unsigned int map_size;
};

spe_program_handle_t
*spe_open_image(const char *filename)
{
	/* allocate an extra integer in the spe handle to keep the mapped size information */
	struct image_handle *ret;
	int binfd = -1, f_stat;
	struct stat statbuf;
	size_t ps = getpagesize ();

	ret = malloc(sizeof(struct image_handle));
	if (!ret)
		return NULL;

	ret->speh.handle_size = sizeof(spe_program_handle_t);
	ret->speh.toe_shadow = NULL;

	binfd = open(filename, O_RDONLY);
	if (binfd < 0)
		goto ret_err;

	f_stat = fstat(binfd, &statbuf);
	if (f_stat < 0)
		goto ret_err;

	/* Sanity: is it executable ?
	 */
	if(!(statbuf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH)))
	{
		errno=EACCES;
		goto ret_err;
	}
	
	/* Sanity: is it small enough ?
	 */
	if(statbuf.st_size > LS_SIZE)
	{
		errno=EACCES;
		goto ret_err;
	}

	/* now store the size at the extra allocated space */
	ret->map_size = (statbuf.st_size + ps - 1) & ~(ps - 1);

	ret->speh.elf_image = mmap(NULL, ret->map_size,
			           PROT_WRITE | PROT_READ,
				   MAP_PRIVATE, binfd, 0);
	if (ret->speh.elf_image == MAP_FAILED)
		goto ret_err;

	/*Verify that this is a valid SPE ELF object*/

	if((verify_spe_elf_image((spe_program_handle_t *)ret)))
		goto ret_err;

	/* ok */
	close(binfd);
	return (spe_program_handle_t *)ret;

	/* err & cleanup */
ret_err:
	if (binfd >= 0)
		close(binfd);

	free(ret);
	return NULL;
}

int
spe_close_image(spe_program_handle_t *handle)
{
	int ret = 0;
	struct image_handle *ih;

	if (!handle)
	{
		errno = EINVAL;
		return -1;
	}
	
	ih = (struct image_handle *)handle;

	if (!ih->speh.elf_image || !ih->map_size)
	{
               	errno = EINVAL;
                return -1;

	}
	
	ret = munmap(ih->speh.elf_image, ih->map_size );
	free(handle);

	return ret;
}

void
register_handler(void * handler, unsigned int callnum ) 
{
	handlers[callnum]=handler;
}

int
check_priority(int policy, int priority)
{
	
	switch (policy) {
	case SCHED_RR:
		if (1 > priority || priority > 99)
		{
			return 0;
		}
		break;
	case SCHED_FIFO:
		if (1 > priority || priority > 99)
		{
			return 0;
		}
		break;
	case SCHED_OTHER:
		if (0 > priority || priority > 40)
		{
			return 0;
		}
		break;
	}

	return 1;
}

/**
 * Library API
 * 
 */

spe_gid_t
spe_create_group (int policy, int priority, int use_events)
{
	struct group_store *group_store;
	
	/* Sanity: check for vaild policy/priority combination.
	 */
	if (!check_priority(policy, priority))
	{
		errno=EINVAL;
		return NULL;
	}

	DEBUG_PRINTF ("spu_create_group(0x%x, 0x%x, 0x%x)\n",
			                  policy, priority, use_events);
	
	group_store=spe_gid_setup(policy, priority, use_events);
	DEBUG_PRINTF("gid is %p\n", group_store);

	return group_store;
}

int
spe_group_max (spe_gid_t spe_gid)
{
	DEBUG_PRINTF ("spu_group_max(0x%x)\n", spe_gid);

	return MAX_THREADS_PER_GROUP;
}


static void *
spe_thread (void *ptr)
{
	/* If the SPU_INFO (or SPU_DEBUG_START) environment variable is set,
	   output a message to stderr telling the user how to attach a debugger
	   to the new SPE thread.  */
	if (__env_spu_debug_start || __env_spu_info)
	  {
		int tid = syscall (__NR_gettid);
		fprintf (stderr, "Starting SPE thread %p, to attach debugger use: spu-gdb -p %d\n",
			 ptr, tid);

		/* In the case of SPU_DEBUG_START, actually wait until the user *has*
		   attached a debugger to this thread.  This is done here by doing an
		   sigwait on the empty set, which will return with EINTR after the
		   debugger has attached.  */
		if (__env_spu_debug_start)
		  {
			sigset_t set;
			sigemptyset (&set);
			/* Use syscall to avoid glibc looping on EINTR.  */
			syscall (__NR_rt_sigtimedwait, &set, (void *) 0,
				 (void *) 0, _NSIG / 8);
		  }
	  }

	return (void *) (unsigned long) do_spe_run (ptr);
}

speid_t
spe_create_thread (spe_gid_t gid, spe_program_handle_t *handle, void *argp, void *envp, unsigned long mask, int flags)
{
	struct thread_store *thread_store;
	int rc;

	/* Sanity check
	 */
	if (!handle || (( flags & SPE_USER_REGS ) && !argp ))
	{
		errno=EINVAL;
		return NULL;
	}

	if ( gid != NULL && !srch_group(gid))
	{
		return NULL;
	}

	if ( mask == 0 ||  (mask & ((1<<spe_group_max(NULL))-1)) == 0 )
	{
		errno=EINVAL;
		return NULL;
	}
	
	DEBUG_PRINTF ("spe_create_thread(%p, %p, %p, %p, 0x%x, 0x%x)\n",
		      gid, handle, argp, envp, mask, flags);

	if (check_env)
	  env_check();

	thread_store = spe_setup (gid, handle, argp, envp, flags);
	if (!thread_store)
	  {
		  DEBUG_PRINTF ("spe_setup failed\n");
		  errno=EINVAL;
		  return NULL;
	  }


	// remember the affinity mask
	//
	thread_store->mask = mask;

#ifndef HAVE_GLIBC23
	DEBUG_PRINTF ("pthread_attr_init()\n");
	rc = pthread_attr_init(&thread_store->thread_attr);
	if (rc)
	{
		perror ("pthread_attr_init");
		return NULL;
	}

	DEBUG_PRINTF ("pthread_attr_setaffinity_np()\n");
	rc = pthread_attr_setaffinity_np(&thread_store->thread_attr, sizeof(unsigned long), (cpu_set_t*)&mask);
	if (rc)
	{
		perror ("pthread_attr_setaffinity_np()");
		return NULL;
	}
#endif	
	DEBUG_PRINTF ("pthread_create()\n");
	rc = pthread_create (&thread_store->spe_thread, NULL,
			     spe_thread, thread_store);

	if (rc)
	  {
		  perror ("pthread_create");
		  return NULL;
	  }

	return thread_store;
}

static inline int spe_wait_status(unsigned int spe_status_R)
{
	unsigned int rc, term = 0;

	rc = (spe_status_R & 0xffff0000) >> 16;
	if (spe_status_R & 0x4) {
		/* SPE halted. */
		term = SIGABRT;
	} else if (spe_status_R & 0x20) {
		/* SPE invalid instruction. */
		term = SIGILL;
	} else if (spe_status_R & 0x40) {
		/* SPE invalid ch. */
		term = SIGILL;
	} else if ((rc < 0x2000) || (rc > 0x20ff)) {
		/* Treat these as invalid stop codes. */
		term = SIGILL;
	} else if (spe_status_R & 0x1) {
		/* Thread killed for undetermined reason. */
		term = SIGKILL;
	}
	/* Return status that can be evaluated with WIFEXITED(), etc. */
	return ((rc & 0xff) << 8) | (term & 0x7f);
}

int 
spe_wait(speid_t speid, int *status, int options)
{
	struct thread_store *thread_store = speid;
	void *pthread_rc;
	int rc;

	if (!srch_thread(speid))
	{
		return -1;
	}	
	
	DEBUG_PRINTF("spu_wait(0x%x, %p, 0x%x)\n", speid, status, options);
/*
	if (options & WNOHANG)
	{
		if (thread_store->->thread_status != 99)
		{
			
		}
	}
*/	
	if (options & WUNTRACED)
	{
		if (thread_store->thread_status == 10 || thread_store->thread_status == 20 )
		{
			*status = 0x4;
		}
	}
	
	rc = pthread_join(thread_store->spe_thread, &pthread_rc);
	if (rc) {
		DEBUG_PRINTF("  pthread_join failed, errno=%d\n", errno);
		return -1;
	}
	if (status) {
		*status = spe_wait_status(thread_store->ret_status);
	}
	spe_cleanup(thread_store);

	DEBUG_PRINTF("Thread ended.\n");
	return rc;
}

int
spe_kill (speid_t speid, int sig)
{
	struct thread_store *thread_store = speid;

	if (!srch_thread(speid))
	{
		return -1;
	}	
	
	if (sig == SIGCONT)
	{
		int stopped = 0;
		
		pthread_mutex_lock(&thread_store->event_lock);
		stopped = thread_store->event_pending;
		pthread_mutex_unlock(&thread_store->event_lock);
		if ( thread_store->stop || stopped )
		{
			pthread_cond_signal(&thread_store->event_deliver);
		}
		return 0;
	}
	if (sig == SIGKILL)
	{	
		int stopped = 0;
		
		/* Tell the thread that it is to be killed.*/
		thread_store->killed = 1;

		/* See if it is halted and waiting */
		pthread_mutex_lock(&thread_store->event_lock);
		stopped = thread_store->event_pending;
		pthread_mutex_unlock(&thread_store->event_lock);
		if (stopped)
		{
			pthread_cond_signal(&thread_store->event_deliver);
			return 0;
	        }
		else 
		{
			return pthread_cancel (thread_store->spe_thread);
		}
	}
	if (sig == SIGSTOP)
	{	
		/* Tell the thread that it is to stop.*/
		thread_store->stop = 1;

		return 0;
	}

	errno = EINVAL;
	return -1;
}

int
validatefd(struct poll_helper *phelper, int pos, int fd)
{
	int i;
	
	for(i=0;i<pos;i++)
	{
		if (phelper[i].retfd == fd)
		{
			DEBUG_PRINTF("Removed double fd. (%i)\n ",fd);
			return 0;
		}
	}
	return 1;
}

int
spe_get_event(struct spe_event *pevents, int nevents, int timeout)
{
	int i;
	int ret_events = 0;
	int numSPEsToPoll = 0;
	int setupSPEs =0;
	int pollRet = 0;

	struct pollfd *SPEfds;

	struct poll_helper *phelper;
	
	pthread_mutex_lock(&grp_list.mutex);

	for (i=0 ; i < nevents ; i++)
	{
		// Clear output fields.
		pevents[i].speid = NULL;
		pevents[i].revents=0;
		pevents[i].data=0;
	}
	
	for (i=0 ; i < nevents ; i++)
	{
		int j;
		struct group_store *group = pevents[i].gid;
		struct grpListElem *elem = group->grp_members;

		DEBUG_PRINTF("spe_get_event():\n");
		DEBUG_PRINTF("  using group  : %p\n", group);
		DEBUG_PRINTF("  with members : %i\n", group->numListSize);

		for ( j=0 ; j < group->numListSize; j++)
		{
			struct thread_store *thread = elem->thread;
		
			DEBUG_PRINTF("  scan member  : %p\n", thread);
			
			// Scan for present events in the thread structures of SPEs
			if (thread->event != 0 && thread->event & pevents[i].events)
			{
				int pipeval,ret;
			
				DEBUG_PRINTF("  has event !  : 0x%04x\n", thread->event);
				
				//Found an event we're looking for.
				ret_events++;

				//Fill out return struct.
				pevents[i].revents = thread->event;
				pevents[i].speid   = thread;
				pevents[i].data    = thread->ev_data;

				//Empty the pipe
				ret=read(thread->ev_pipe[0], &pipeval, 4);
				thread->ev_data = 0;
				thread->event = 0;

				pthread_mutex_unlock(&grp_list.mutex);
				return ret_events;
			}
			
			//Decide on what fd's to poll on
			if (pevents[i].events & (SPE_EVENT_MAILBOX))
			{
				numSPEsToPoll++;
			}
			if (pevents[i].events & (SPE_EVENT_TAG_GROUP))
			{
				numSPEsToPoll++;
			}
			if (pevents[i].events & (SPE_EVENT_STOP | SPE_EVENT_DMA_ALIGNMENT |
						SPE_EVENT_SPE_ERROR | SPE_EVENT_SPE_DATA_SEGMENT | 
						SPE_EVENT_SPE_DATA_STORAGE | SPE_EVENT_SPE_TRAPPED | SPE_EVENT_THREAD_EXIT ))
			{
				numSPEsToPoll++;
			
			}
			elem=elem->next;
		}
	}
	
	if(numSPEsToPoll == 0)
	{
		pthread_mutex_unlock(&grp_list.mutex);
		errno=EINVAL;
		return -1;
	}

	if (ret_events > 0)
	{
		//printf("P1\n");
		
		//DEBUG_PRINTF("  returning !  : 0xi\n", ret_events);

		pthread_mutex_unlock(&grp_list.mutex);
		
		return ret_events;
	}
	
	DEBUG_PRINTF("  number of fd : %i\n", numSPEsToPoll);

	SPEfds=malloc (numSPEsToPoll * sizeof(struct pollfd));
	phelper=malloc (numSPEsToPoll * sizeof(struct poll_helper));

	// Set up all necessary fds to poll on and remeber what they are for.
	for (i=0 ; i < nevents ; i++)
	{
		int j;
		struct group_store *group = pevents[i].gid;
		struct grpListElem *elem = group->grp_members;

		for ( j=0 ; j < group->numListSize; j++)
		{
			struct thread_store *thread = elem->thread;
			
			if (pevents[i].events & (SPE_EVENT_MAILBOX))
			{
				SPEfds[setupSPEs].fd=thread->fd_ibox;
				SPEfds[setupSPEs].events=POLLIN;
			
				phelper[setupSPEs].event=i;
				phelper[setupSPEs].thread=thread;
				phelper[setupSPEs].type=1;
				//printf("1\n");
				setupSPEs++;
			}
			if (pevents[i].events & (SPE_EVENT_STOP | SPE_EVENT_DMA_ALIGNMENT |
						SPE_EVENT_SPE_ERROR | SPE_EVENT_SPE_DATA_SEGMENT | 
						SPE_EVENT_SPE_DATA_STORAGE | SPE_EVENT_SPE_TRAPPED | SPE_EVENT_THREAD_EXIT ))
			{
				SPEfds[setupSPEs].fd=thread->ev_pipe[0];
				SPEfds[setupSPEs].events=POLLIN;
			
				phelper[setupSPEs].event=i;
				phelper[setupSPEs].thread=thread;
				phelper[setupSPEs].type=2;
				//printf("2\n");
				setupSPEs++;
			}
			if (pevents[i].events & (SPE_EVENT_TAG_GROUP))
			{
				SPEfds[setupSPEs].fd=thread->fd_mfc;
				SPEfds[setupSPEs].events=POLLIN;

				phelper[setupSPEs].event=i;
				phelper[setupSPEs].thread=thread;
				phelper[setupSPEs].type=3;
				if (SPEfds[setupSPEs].fd == -1)
					fprintf(stderr, "Warning: spe_get_events: attempting "
						"to wait for tag group without DMA support\n");
				//printf("3\n");
				setupSPEs++;
			}
			elem=elem->next;
		}
	}
	
	if (setupSPEs != numSPEsToPoll)
	{
		DEBUG_PRINTF("ERROR:Thread number mismatch.");
		
		free(SPEfds);
		free(phelper);
		
		pthread_mutex_unlock(&grp_list.mutex);
		errno = EFAULT;
		return -1;
	}

	pthread_mutex_unlock(&grp_list.mutex);
	
	DEBUG_PRINTF("Polling for %i fd events.\n",setupSPEs);

	pollRet = poll(SPEfds, setupSPEs, timeout);
	DEBUG_PRINTF("Poll returned %i events.\n",pollRet);
	
	ret_events = 0;
	
	// Timeout.
	if (pollRet == 0)
	{
		free(SPEfds);
		free(phelper);
		return 0;
	}
	
	
	for (i=0 ; i < setupSPEs ; i++ )
	{
		if (SPEfds[i].revents != 0 && pevents[phelper[i].event].speid == NULL )
		{
			int rc,data;
			
			if (validatefd(phelper,i,SPEfds[i].fd))
			{	
				switch (phelper[i].type) {
				case 1:
					// Read ibox data if present
	
					rc = read(SPEfds[i].fd, &data, 4);
					if (rc == 4)
					{
						phelper[i].retfd=SPEfds[i].fd;
						pevents[phelper[i].event].data = data;
						pevents[phelper[i].event].speid = phelper[i].thread;
						pevents[phelper[i].event].revents = SPE_EVENT_MAILBOX;
						ret_events++;
					}
					break;
				case 2:
       	                        	// Read pipe data if present
	
       	                        	rc = read(SPEfds[i].fd, &data, 4);
       	                        	//pevents[phelper[i].event].revents = pevents[phelper[i].event].revents | data;
					phelper[i].retfd=SPEfds[i].fd;
       	                        	pevents[phelper[i].event].revents = data;
					pevents[phelper[i].event].speid = phelper[i].thread;
				       	pevents[phelper[i].event].data = phelper[i].thread->ev_data;
	
				       	phelper[i].thread->ev_data = 0;
				       	phelper[i].thread->event = 0;
					ret_events++;
					break;
				case 3:
					// Read tag group data if present
	
					rc = read(SPEfds[i].fd, &data, 4);
					if (rc == 4)
					{	
						phelper[i].retfd=SPEfds[i].fd;
						pevents[phelper[i].event].data = data;
						pevents[phelper[i].event].speid = phelper[i].thread;
						pevents[phelper[i].event].revents = SPE_EVENT_TAG_GROUP;
						ret_events++;
					}
					break;
				}
			}
		}
	}
	
	free(SPEfds);
	free(phelper);
	//printf("P2\n");
	return ret_events;
}

int
spe_get_threads(spe_gid_t gid, speid_t *spe_ids)
{
	struct group_store *group = gid;
	int i;
	
	if (!spe_ids)
	{
		return group->numListSize;
	}
	else
	{
		struct grpListElem *elem = group->grp_members;
		
		for(i=0; i < group->numListSize ; i++)
		{
			spe_ids[i] = elem->thread;
			elem=elem->next;
			
		}
	}

	return i;
}

void
*spe_get_ls (speid_t speid)
{
	struct thread_store *thread_store = speid;
	
	if (!srch_thread(speid))
	{
		return NULL;
	}
	
#ifdef SPE_NO_MMAP
	errno = EFAULT;
	return -1;
#else
	return thread_store->mem_mmap_base;
#endif
}

int spe_get_priority(spe_gid_t gid)
{
	struct group_store *group_store = gid;
	
	return group_store->priority;
}

int spe_set_priority(spe_gid_t gid, int priority)
{
	struct group_store *group_store = gid;

	/* Sanity: check for vaild policy/priority combination.
        */
	if (!check_priority(group_store->policy, priority))
        {
                errno=EINVAL;
                return -1;
        }

	
	group_store->priority=priority;

	/*int pthread_setschedparam(pthread_t target_thread,  int  policy,  const struct sched_param *param);*/
	return 0;
}

int spe_get_policy(spe_gid_t gid)
{
	struct group_store *group_store = gid;
	
	return group_store->policy;
}

spe_gid_t
spe_get_group (speid_t speid)
{
	struct thread_store *thread_store = speid;

	if (!srch_thread(speid))
	{
		return NULL;
	}	
	
	return thread_store->group_id;
}

int
spe_get_affinity( speid_t speid, unsigned long *mask)
{
	struct thread_store *thread_store = speid;
	int result = 0;

	if (!srch_thread(speid))
	{
		return -1;
	}	
#ifdef HAVE_GLIBC23	
	result = sched_getaffinity(thread_store->threadid, sizeof(unsigned long), mask);
#else 
	result = pthread_getaffinity_np(thread_store->spe_thread, sizeof(unsigned long), (cpu_set_t*)mask);
#endif
	return result;
}

int
spe_set_affinity(speid_t speid, unsigned long mask)
{
	struct thread_store *thread_store = speid;
	int result = 0;

	if (!srch_thread(speid))
	{
		return -1;
	}	
	thread_store->mask = mask;

#ifdef HAVE_GLIBC23
	result = sched_setaffinity(thread_store->threadid, sizeof(unsigned long), &mask);
#else
	result = pthread_setaffinity_np(thread_store->spe_thread, sizeof(unsigned long), (cpu_set_t*)&mask);
#endif
	return result;
}

int
spe_get_context(speid_t speid, struct spe_ucontext *uc)
{
	printf("spe_get_context: not implemented in this release.\n");

	errno=ENOSYS;
	return -1;
}

int
spe_set_context(speid_t speid, struct spe_ucontext *uc)
{
	printf("spe_set_context: not implemented in this release.\n");

	errno=ENOSYS;
	return -1;
}

int 
spe_group_defaults(int policy, int priority, int spe_events)
{
        /* Sanity: check for vaild policy/priority combination.
	 */
	
	if (!check_priority(policy, priority))
	{
		errno=EINVAL;
		return -1;
	}

	default_policy = policy;
	default_priority = priority;
	default_eventmask= spe_events;

	return 0;
}

int __spe_get_context_fd(speid_t speid)
{
	struct thread_store *thread_store = speid;
	
	if (!srch_thread(speid))
	{
		return -1;
	}	

	return thread_store->fd_run;
}
		

/*
 * mfc.h direct call-ins
 *
 * */

int spe_write_in_mbox (speid_t spu_ps_addr ,unsigned int data)
{
	struct thread_store *thread_store = spu_ps_addr;
	int rc;
	
	if (!srch_thread(spu_ps_addr))
	{
		return -1;
	}	
	
	if (thread_store->thread_status == 99)
		return -1;

	rc = write(thread_store->fd_wbox, &data, 4);
	if (rc == 4)
		rc = 0;

	return rc;

}

int spe_stat_in_mbox (speid_t spu_ps_addr)
{
	struct thread_store *thread_store = spu_ps_addr;
	int rc, ret;

	if (!srch_thread(spu_ps_addr))
	{
		return -1;
	}	
	
	rc = read(thread_store->fd_wbox_stat, &ret, 4);
	if (rc != 4)
		ret = -1;

	return ret;

}

unsigned int spe_read_out_mbox (speid_t spu_ps_addr)
{
        struct thread_store *thread_store = spu_ps_addr;
        int rc;
	unsigned int ret;

	if (!srch_thread(spu_ps_addr))
	{
		return 0;
	}	
	
        rc = read(thread_store->fd_mbox, &ret, 4);
        if (rc != 4)
	        ret = 0;

        return ret;

}


int spe_stat_out_mbox (speid_t spu_ps_addr)
{
	struct thread_store *thread_store = spu_ps_addr;
        int rc, ret;

	if (!srch_thread(spu_ps_addr))
	{
		return -1;
	}	
	
        rc = read(thread_store->fd_mbox_stat, &ret, 4);
        if (rc != 4)
                ret = -1;

        return ret;

}


int spe_stat_out_intr_mbox (speid_t spu_ps_addr)
{
        struct thread_store *thread_store = spu_ps_addr;
        int rc, ret;

	if (!srch_thread(spu_ps_addr))
	{
		return -1;
	}	
	
        rc = read(thread_store->fd_ibox_stat, &ret, 4);
        if (rc != 4)
                ret = -1;

        return ret;

}


int spe_write_signal (speid_t spu_ps_addr, unsigned int signal_reg, unsigned int data)
{
	struct thread_store *thread_store = spu_ps_addr;
	int rc;
	
	if (!srch_thread(spu_ps_addr))
	{
		return -1;
	}	
	
	if (thread_store->thread_status == 99)
		return -1;

	if (signal_reg == SPE_SIG_NOTIFY_REG_1)
		rc = write(thread_store->fd_sig1, &data, 4);
	else if (signal_reg == SPE_SIG_NOTIFY_REG_2)
		rc = write(thread_store->fd_sig2, &data, 4);
	else
		return -1;
	
	if (rc == 4)
		rc = 0;

	return rc;
}

